# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
This function imports a Viva Insights Query from a .csv file and optimizes the
variable classifications for other functions in the package. The function takes in a file path (x)
and an optional encoding parameter (default is 'utf-8'). It checks if the file is a .csv file, reads
in the file using pandas, cleans the column names by removing spaces and special characters, and
returns the resulting data as a pandas dataframe. If there is an error reading the file, the function prints an error message.
"""
import pandas as pd
import re 
import os

def import_query(x, encoding: str = 'utf-8'):
    """
    Name
    ----
    import_query

    Description
    ----------- 
    The function `import_query` reads a CSV file, removes leading and trailing spaces from column names,
    and replaces spaces and special characters with underscores in column names.
    
    Parameters
    ----------
    x : str
        The parameter `x` is the input file name or path. It should be a string representing the file name or path of the CSV file you want to import.
    encoding : str, optional
        The encoding parameter specifies the character encoding to be used when reading the CSV file. The default value is 'utf-8', which is a widely used encoding for text files. However, you can specify a different encoding if needed.
    
    Returns
    -------
    The variable `data` if the input file is a valid CSV file. If the input file is not a valid CSVfile, the function will print an error message and return `None`.
    """    
    
    # in case '.csv' is not all in lower case, make it lower case
    if x[-4:].lower() == '.csv':
        in_df = x[:-4] + '.csv'
    else:
        in_df = x
        
    if not os.path.isfile(in_df):
        raise ValueError("input file does not exist")
    
    elif not in_df.endswith('.csv'):
        raise ValueError("the input must be a .csv file")
    
    else:
        try:
            # Try to read in csv file, if file can not be read, exception is thrown.
            data = pd.read_csv(x, encoding=encoding, delimiter=',')
            # Replace mentions of '%' with literal string 'Percent'
            data.columns = [re.sub('%', 'Percent', c.strip()) for c in data]
            # Remove leading and trailing spaces
            # Remove spaces and special characters and replacing them with underscores within column names.
            data.columns = [re.sub('[^a-zA-Z0-9,]', '_', c.strip()) for c in data]
            
            return data
        except: 
            raise ValueError('something went wrong when reading the file')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
The code defines a function `create_bar` that calculates and visualizes the mean of a selected
metric, grouped by a selected HR variable. 

The metrics are first aggregated at a user-level prior to being aggregated at the level of the HR variable. The function `create_bar` returns either a plot object or a table, depending on the value passed to `return_type`. 
"""
import pandas as pd
import seaborn as sns
from vivainsights.extract_date_range import extract_date_range
from vivainsights.us_to_space import us_to_space
import matplotlib.ticker as mtick
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedLocator
import matplotlib
    
def create_bar_calc(
    data: pd.DataFrame,
    metric: str,
    hrvar: str, 
    mingroup = 5
    ):
    """Calculate the mean of a selected metric, grouped by a selected HR variable."""
    data = data.groupby(['PersonId',hrvar])        
    data = data[metric].mean()
    data = data.reset_index()
    output = data.groupby(hrvar).agg(
        metric = (metric, 'mean'),
        n = ('PersonId', 'nunique')
        )
    output = output[output['n'] >= mingroup]
    output = output.rename_axis(hrvar).reset_index()
    output = output.sort_values(by = 'metric', ascending=False)
    return output

def create_bar_viz(
    data: pd.DataFrame,
    metric: str,
    hrvar: str,
    mingroup = 5,
    percent: bool = False,
    plot_title = None,
    plot_subtitle = None):
    """Visualise the mean of a selected metric, grouped by a selected HR variable."""
    sum_df = create_bar_calc(data, metric, hrvar, mingroup)
    caption_text = extract_date_range(data, return_type='text')
    plot_order = sum_df[hrvar].to_numpy()

    # Title and subtitle text
    if plot_title is None:
        title_text = us_to_space(metric)
    else:
        title_text = plot_title

    if plot_subtitle is None:
        subtitle_text = f'Weekly average by {hrvar}'  # TODO: make this dynamic by date interval
    else:
        subtitle_text = plot_subtitle

    # fig = plt.figure()
    fig, ax = plt.subplots(figsize=(4, 6))

    # Create grid
    # Zorder tells it which layer to put it on. We are setting this to 1 and our data to 2 so the grid is behind the data.
    ax.grid(which="major", axis='x', color='#758D99', alpha=0.6, zorder=1)

    # Remove splines. Can be done 1 at a time or can slice with a list.
    ax.spines[['top', 'right', 'bottom']].set_visible(False)

    # Make left spine slightly thicker
    ax.spines['left'].set_linewidth(1.1)

    ax.barh(sum_df[hrvar], sum_df['metric'], color='#1d627e', zorder=2)

    if percent == True:
        # Set the x-axis format to percentage
        ax.xaxis.set_major_formatter(mtick.PercentFormatter(xmax=1.0))

    # Shrink y-lim to make plot a bit tighter
    # Using length of summary table to make it dynamic
    ax.set_ylim(-0.5, len(sum_df) - 0.5)

    # Reformat x-axis tick labels
    ax.xaxis.set_tick_params(labeltop=True,  # Put x-axis labels on top
                             labelbottom=False,  # Set no x-axis labels on bottom
                             bottom=False,  # Set no ticks on bottom
                             labelsize=9,  # Set tick label size
                             pad=-1)  # Lower tick labels a bit

    ax.yaxis.set_tick_params(pad=10,  # Pad tick labels so they don't go over y-axis
                             labelsize=9,  # Set label size
                             bottom=False)  # Set no ticks on bottom/left

    # Reformat y-axis tick labels
    ax.set_yticks(range(len(sum_df)))
    ax.set_yticklabels(sum_df[hrvar], ha='right')

    # Add in line and tag
    ax.plot([-.35, .87],  # Set width of line
            [1.02, 1.02],  # Set height of line
            transform=fig.transFigure,  # Set location relative to plot
            clip_on=False,
            color='#fe7f4f',
            linewidth=.6)

    ax.add_patch(plt.Rectangle((-.35, 1.02),  # Set location of rectangle by lower left corder
                               0.12,  # Width of rectangle
                               -0.02,  # Height of rectangle. Negative so it goes down.
                               facecolor='#fe7f4f',
                               transform=fig.transFigure,
                               clip_on=False,
                               linewidth=0))

    # Add in title, subtitle, and caption
    ax.text(x=-.35, y=.96, s=title_text, transform=fig.transFigure, ha='left', fontsize=13, weight='bold', alpha=.8)
    ax.text(x=-.35, y=.925, s=subtitle_text, transform=fig.transFigure, ha='left', fontsize=11, alpha=.8)
    ax.text(x=-.35, y=.08, s=caption_text, transform=fig.transFigure, ha='left', fontsize=9, alpha=.7)

    if percent == True:
        ax.bar_label(ax.containers[0], labels=[f"{100 * value:.0f}%" for value in sum_df['metric']], label_type="edge",
                     padding=3)
    else:
        ax.bar_label(ax.containers[0], fmt='%.0f', label_type='edge', padding=3)  # annotate

    ax.margins(y=0.3)  # pad the spacing between the number and the edge of the figure

    # return the plot object
    return fig    

def create_bar(
    data: pd.DataFrame,
    metric: str,
    hrvar: str,
    mingroup = 5,
    percent: bool = False,
    return_type: str = "plot",
    plot_title = None,
    plot_subtitle = None):
    """
    Name
    -----
    create_bar 
    
    Description
    -----------
    The function `create_bar` calculates and visualizes the mean of a selected metric, grouped by a selected HR variable. The metrics are first aggregated at a user-level prior to being aggregated at the level of the HR variable. The function `create_bar` returns either a plot object or a table, depending on the value passed to `return_type`.

    Parameters
    ----------
    data : pandas dataframe
        person query data
    metric : str
        name of the metric to be analysed
    hrvar : str
        name of the organizational attribute to be used for grouping

    Returns
    -------
    _type_
        the type of output to return. Defaults to "plot".

    Example
    -------
    >>> create_bar(pq_data, metric = "Collaboration_hours", hrvar = "LevelDesignation")
    """  
    if return_type == "plot":
        out = create_bar_viz(data=data, metric=metric, hrvar=hrvar, percent=percent, mingroup=mingroup, plot_title = plot_title, plot_subtitle = plot_subtitle)
    elif return_type == "table":
        out = create_bar_calc(data=data, metric=metric, hrvar=hrvar, mingroup=mingroup)
    else:
        out = "Invalid input. Please check your inputs and try again."
    return out


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
The `create_trend` function provides a week by week view of a selected Viva Insights metric, 
allowing you to either return a week by week heatmap bar plot or a summary table. 
By default, `create_trend` returns a week by week heatmap bar plot, highlighting the points intime with most activity. 
Additional options available to return a summary table.
"""

import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import seaborn as sns
from vivainsights.extract_date_range import extract_date_range

def create_trend(data: pd.DataFrame,
  
                 metric: str,
                 palette = [
                    "#0c3c44",
                    "#1d627e",
                    "#34b1e2",
                    "#bfe5ee",
                    "#fcf0eb",
                    "#fbdacd",
                    "#facebc",
                    "#fe7f4f"
                    ],
                 hrvar: str = "Organization",
                 mingroup = 5,
                 return_type: str = "plot",
                 legend_title: str = "Hours",
                 date_column: str = "MetricDate",
                 date_format: str = "%Y-%m-%d"):  
  """
  Name
  ----
  create_trend

  Description
  -----------
  This module provides a week by week view of a selected Viva Insights metric. 
  By default returns a week by week heatmap bar plot, highlighting the points intime with most activity. 
  Additional options available to return a summary table.
    
  Parameters
  ---------
  data : panda dataframe
      The input data as a pandas DataFrame.
  metric : str
    The metric parameter is a string that represents the column name in the data DataFrame that contains the values to be plotted or analyzed. This could be any numerical metric such as sales, revenue, or number of hours worked.
  palette : list
    The `palette` parameter is a list of colors that will be used to represent different groups in the trend plot. Each color in the list corresponds to a different group. By default, the palette includes 8 colors, but you can modify it to include more or fewer colors if needed.
  hrvar : str
    hrvar is a string parameter that represents the variable used for grouping the data. In this case, it is used to group the data by organization. Defaults to Organization mingroup: The `mingroup` parameter is used to specify the minimum number of groups that should be present in the data for the trend analysis. If the number of unique values in the `hrvar` column is less than `mingroup`, the function will raise an error. Defaults to 5
  return_type : str
      The `return_type` parameter determines the type of output that the function will return. It can have two possible values:. Defaults to plot
  legend_title : str
      The title for the legend in the plot. It is used to label the different categories or groups in the data. Defaults to Hours
  date_column : str
      The name of the column in the DataFrame that contains the dates for the trend analysis. Defaults to MetricDate
  date_format : str
      The `date_format` parameter is used to specify the format of the dates in the `date_column` of the input data. It should be a string that follows the syntax of the Python `datetime` module's `strftime` function. This allows you to specify how the dates are formatted in the. Defaults to %Y-%m-%d

  Returns
  ------
  The function `create_trend` returns either a table or a plot, depending on the value of the `return_type` parameter.

  Example
  -------
  >>> create_trend(data = pq_data, metric = "Collaboration_hours", hrvar = "LevelDesignation")
  """
  # Return the table or the plot or raise an error
  if return_type == "table":    
    myTable = create_trend_calc(data, metric, hrvar, mingroup, date_column, date_format)
    myTable_return = myTable.pivot(index="group", columns=date_column, values=metric)
    return myTable_return    
  elif return_type == "plot":
    return create_trend_viz(data, metric, palette, hrvar, mingroup, legend_title, date_column, date_format)
  else:
    raise ValueError("Please enter a valid input for return_type.")
  
def create_trend_calc(data, metric, hrvar, mingroup, date_column, date_format):
  """
  Name 
  ----
  create_trend_calc

  Description
  -----------
  This function creates a trend calculation by grouping data by a specified variable and calculating
  the mean of a specified metric over time.
  """
  # Check inputs
  required_variables = [date_column, metric, "PersonId"]

  # Error message if variables are not present
  # Nothing happens if all present
  for var in required_variables:
    if var not in data.columns:
      raise ValueError(f"{var} is not in the data")

  # Handling None values passed to hrvar
  if hrvar is None:
    data["Total"] = 1 # Create a dummy column for totals
    hrvar = "Total"

  # Clean metric name
  clean_nm = metric.replace("_", " ")

  # Convert Date to datetime and rename hrvar to group
  data[date_column] = pd.to_datetime(data[date_column], format=date_format)
  data = data.rename(columns={hrvar: "group"})

  # Select relevant columns and group by group
  myTable = data[["PersonId", date_column, "group", metric]]
  myTable = myTable.groupby("group")

  # Calculate employee count and filter by mingroup
  myTable = myTable.apply(lambda x: x.assign(Employee_Count = x["PersonId"].nunique()))
  myTable = myTable[myTable["Employee_Count"] >= mingroup]

  # Group by date and group and calculate mean metric and employee count
  myTable.reset_index(drop = True, inplace = True)
  myTable = myTable.groupby([date_column, "group"]).agg({"Employee_Count": "mean", metric: "mean"}).reset_index()

  return myTable
  
  
def create_trend_viz(
  data: pd.DataFrame,
  metric: str,
  palette,
  hrvar: str,
  mingroup,
  legend_title: str,
  date_column: str,
  date_format: str
):
  """
  Name
  ----
  create_trend_viz

  Description
  -----------
  This function creates a heatmap visualization of trends in a given metric by a specified variable
  over time.
  """
  
  myTable = create_trend_calc(data, metric, hrvar, mingroup, date_column, date_format)
  myTable_plot = myTable[[date_column, "group", metric]]  
  # myTable_plot[date_column] = pd.to_datetime(myTable[date_column], format=date_format)
  # myTable_plot[date_column] = pd.to_datetime(myTable[date_column], format=date_format).dt.date
  
  # Clean labels for plotting
  clean_nm = metric.replace("_", " ")  
  title_text = f"{clean_nm}\nHotspots by {hrvar.lower()}"
  subtitle_text = f'By {hrvar}'
  caption_text = extract_date_range(data, return_type = 'text')  
  
  # Create the plot object
  # Setup plot size.
  fig, ax = plt.subplots(figsize=(7, 4))
    
  # Remove tick marks
  ax.tick_params(
      which='both',      # Both major and minor ticks are affected
      top=False,         # Remove ticks from the top
      bottom=False,      # Remove ticks from the bottom
      left=False,        # Remove ticks from the left
      right=False        # Remove ticks from the right
  )
  
  # Create heatmap plot
  sns.heatmap(
    data = myTable_plot.pivot(index="group", columns=date_column, values=metric),
    cmap = palette,
    cbar_kws={"label": legend_title},
    xticklabels= myTable_plot[date_column].dt.date.unique()
    )
  
  # Reformat x-axis tick labels
  ax.xaxis.set_tick_params(labelsize = 9, rotation=45)
  ax.yaxis.set_tick_params(labelsize = 9)
  # ax.xaxis.set_major_formatter(mdates.DateFormatter('%d %b %y'))
  
  # Remove axis labels
  ax.set_xlabel('')
  ax.set_ylabel('')
  
  ax.plot(
    [-0.08, .9], # Set width of line, previously [-0.08, .9]
    [0.9, 0.9], # Set height of line
    transform = fig.transFigure, # Set location relative to plot
    clip_on = False,
    color = '#fe7f4f',
    linewidth = .6
)

  ax.add_patch(
      plt.Rectangle(
          (-0.08, 0.9), # Set location of rectangle by lower left corner, previously [-0.08, .9]
          0.05, # Width of rectangle
          -0.025, # Height of rectangle
          facecolor = '#fe7f4f',
          transform = fig.transFigure,
          clip_on = False,
          linewidth = 0
          )
  )        

  # Set title
  ax.text(
      x = -0.08, y = 1.00,
      s = title_text,
      transform = fig.transFigure,
      ha = 'left',
      fontsize = 13,
      weight = 'bold',
      alpha = .8
  )

  # Set subtitle
  ax.text(
      x = -0.08, y = 0.95,
      s = subtitle_text,
      transform = fig.transFigure,
      ha = 'left',
      fontsize = 11,        
      alpha = .8
  )

  # Set caption
  ax.text(x=-0.08, y=-0.12, s=caption_text, transform=fig.transFigure, ha='left', fontsize=9, alpha=.7)
  
  # return the plot object
  return fig
  
  """Legacy 
  # plot_object.set_title(f"{clean_nm}\nHotspots by {hrvar.lower()}")
  # plot_object.set_xlabel(date_column)
  # plot_object.set_ylabel(hrvar)
  """


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
This module returns a data frame containing a meeting query.
"""

import pkg_resources
import pandas as pd

def load_mt_data():
    # This is a stream-like object. If you want the actual info, call
    # stream.read()
    # stream = pkg_resources.resource_string('vivainsights', 'data/mt_data.csv')
    if(pkg_resources.resource_exists(__name__, 'data/mt_data.csv')):
        stream = pkg_resources.resource_stream(__name__, 'data/mt_data.csv')
    elif(pkg_resources.resource_exists(__name__, '../data/mt_data.csv')):
        stream = pkg_resources.resource_stream(__name__, '../data/mt_data.csv')
    else:
        print('Error: please report issue to repo maintainer')    
    return pd.read_csv(stream, encoding='utf-8')


import unittest
import matplotlib.pyplot as plt
from vivainsights.create_bar import create_bar_viz
from vivainsights.pq_data import load_pq_data
import tracemalloc
import gc

def returnfig():
    fig = plt.figure()
    return fig

class TestCreateBarViz(unittest.TestCase):
    def test_create_bar_viz(self):
        tracemalloc.start()        
        pq_data = load_pq_data()        
        fig = create_bar_viz(data = pq_data, metric='Emails_sent', hrvar='Organization')
        
        # garbage collection
        del pq_data
        gc.collect()
        
        tracemalloc.stop()
        self.assertIsInstance(fig, plt.Figure)     

if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
import matplotlib.pyplot as plt
import vivainsights as vi
import warnings
import igraph as ig

class TestNetworkG2G(unittest.TestCase):
    def setUp(self):
        # set up g2g_data
        self.g2g_data = vi.load_g2g_data()

    def test_return_type_table(self):
        # Call the network_g2g function with return_type = "table"
        result = vi.network_g2g(data = self.g2g_data, return_type="table")
        
        # Check if the result is a pandas DataFrame
        self.assertIsInstance(result, pd.DataFrame)

    def test_return_type_data(self):

        with warnings.catch_warnings(record=True) as w: 
            # Call the network_g2g function with return_type = "data"
            result = vi.network_g2g(data = self.g2g_data, return_type="data")
        
            # Check if the result is a pandas DataFrame
            self.assertIsInstance(result, pd.DataFrame)

            # Check if any warnings were generated
            self.assertLess(len(w), 1)

    def test_return_type_network(self):
            
            with warnings.catch_warnings(record=True) as w: 
    
                # Call the network_g2g function with return_type = "network"
                result = vi.network_g2g(data = self.g2g_data, return_type="network")
            
                # Check if the result is an iGraph graph
                self.assertIsInstance(result, ig.Graph)
    
                # Check if any warnings were generated
                self.assertLess(len(w), 1)

    def test_return_type_plot(self):
        with warnings.catch_warnings(record=True) as w: 
            # Call the network_g2g function with return_type = "plot"
            result = vi.network_g2g(data = self.g2g_data, return_type="plot")
        
            # Check if a plot was created
            self.assertIsNotNone(plt.gcf())

            # Check if any warnings were generated
            self.assertLess(len(w), 1)

#run unit tests
if __name__ == '__main__':
    unittest.main()



import unittest
from igraph import Graph
import pandas as pd
from vivainsights import network_summary
import vivainsights as vi
import igraph as ig

class TestNetworkSummaryFunction(unittest.TestCase):
    def setUp(self):
        # Data preparation for network_summary
        data=vi.p2p_data_sim()
        weight = None
        hrvar = "Organization"
        #Set data frame for edges
        if weight is None:
            edges = data.assign(NoWeight = 1).loc[:, ["PrimaryCollaborator_PersonId", "SecondaryCollaborator_PersonId", "NoWeight"]].rename(columns = {"NoWeight": "weight"})
        else:
            edges = data.loc[:, ["PrimaryCollaborator_PersonId", "SecondaryCollaborator_PersonId", weight]]
        pc_hrvar = "PrimaryCollaborator_" + hrvar
        sc_hrvar = "SecondaryCollaborator_" + hrvar
        # TieOrigin = PrimaryCollaborator
        tieOrigin = (
            edges[["PrimaryCollaborator_PersonId"]].drop_duplicates()
            .merge(data[["PrimaryCollaborator_PersonId", pc_hrvar]], on = "PrimaryCollaborator_PersonId", how = "left") #left join
            .rename(columns = {"PrimaryCollaborator_PersonId": "node"})
            .assign(**{hrvar: lambda row: row[pc_hrvar]}) #assign new column
            .drop(columns = [pc_hrvar]) 
        )
        # TieDest = SecondaryCollaborator
        tieDest = (
            edges[["SecondaryCollaborator_PersonId"]].drop_duplicates()
            .merge(data[["SecondaryCollaborator_PersonId", sc_hrvar]], on = "SecondaryCollaborator_PersonId", how = "left")
            .rename(columns = {"SecondaryCollaborator_PersonId": "node"})
            .assign(**{hrvar: lambda row: row[sc_hrvar]})
            .drop(columns = [sc_hrvar])
        )
        # Vertices data frame to provide meta-data
        vert_ft = pd.concat([tieOrigin, tieDest]).drop_duplicates()
        # Create igraph object
        g_raw = ig.Graph.TupleList(edges.itertuples(index=False), directed=True, weights=True)
        # Assign vertex attributes - HR attribute and node
        g_raw.vs[hrvar] = vert_ft[hrvar].tolist()
        g_raw.vs["node"] = vert_ft["node"].tolist()
        # Assign weights
        g_raw.es["weight"] = edges["weight"]
        g=g_raw
        return(g)

    def test_network_summary_table_return(self):
        
        # Test if the function returns a table (pandas DataFrame)
        g=TestNetworkSummaryFunction.setUp(self)
        generated_object= vi.network_summary(g,hrvar = "Organization", return_type="table")
        self.assertIsInstance(generated_object, pd.DataFrame)
        
        # Test if the DataFrame has the expected columns
        expected_columns = ['node_id', 'betweenness', 'closeness', 'degree', 'eigenvector','pagerank', 'hrvar']
        self.assertListEqual(list(generated_object.columns), expected_columns)

        # Test if the DataFrame is not empty
        self.assertFalse(generated_object.empty)

    def test_network_summary_network_return(self):
        
        # Test if the function returns a network (igraph object)
        g=TestNetworkSummaryFunction.setUp(self)
        result_network = network_summary(g, return_type='network')
        self.assertIsInstance(result_network, Graph)

        # Test if node attributes are updated in the graph
        self.assertTrue(all(attr in result_network.vs.attributes() for attr in ['betweenness', 'closeness', 'degree', 'eigenvector', 'pagerank']))

    def test_network_summary_invalid_return_type(self):
        
        # Test if the function raises a ValueError for an invalid return type
        g=TestNetworkSummaryFunction.setUp(self)
        with self.assertRaises(ValueError):
            network_summary(g, return_type='invalid_type')

if __name__ == '__main__':
    unittest.main()


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
Takes a vector of dates and identify whether the frequency is 'daily', 'weekly', or 'monthly'. 
The primary use case for this function is to provide an accurate description of the query type 
used and for raising errors should a wrong date grouping be used in the data input.
"""
import pandas as pd

def identify_datefreq(x):
    x = pd.to_datetime(x)
    
    # Data frame for checking
    date_df = pd.DataFrame({
        "weekdays": pd.Series(list(pd.Series(x).dt.weekday.unique())),
        "n": pd.Series(list(pd.Series(x).dt.weekday.value_counts()))
    })
    
    if len(pd.Series(x).dt.month_name().unique()) == len(x):
        return "monthly"
    elif sum(date_df.loc[date_df.weekdays == 6].n) == len(x): # Check number of Sundays - should equal number of weeks if weekly
        return "weekly"
    elif len(date_df) >= 3: # At least 3 days of the week must be present
        return "daily"
    else:
        return "Unable to identify date frequency."




import unittest
from vivainsights import extract_hr
from vivainsights.pq_data import load_pq_data
import pandas as pd


class TestExtractHr(unittest.TestCase):

    def test_extract_hr(self):
        pq_data=load_pq_data()
        # Call the function to generate the table and text
        generated_table=extract_hr(pq_data, max_unique =50, exclude_constants =True, return_type = "vars")

        # Check if the table is not empty
        self.assertFalse(generated_table.empty)

        generated_text=extract_hr(pq_data, max_unique =50, exclude_constants =True, return_type = "names")
        
        # Check if the text is not empty
        self.assertFalse(generated_text)

if __name__ == '__main__':
    unittest.main()


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
This module creates an incidence analysis reflecting the proportion of the population scoring above or below a specified threshold for a metric. 
"""
import typing
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from vivainsights.color_codes import COLOR_PALLET_ALT_2
from vivainsights.create_bar import create_bar
from vivainsights.extract_date_range import extract_date_range

def create_inc(data: pd.DataFrame, metric: str, hrvar: typing.List or str, mingroup: int = 5, threshold: float = None, position: str = None, return_type: str = 'plot'):
    """
    Name
    ----
    create_inc

    Description
    -----------
    Create an incidence analysis reflecting proportion of population scoring above or below a threshold for a metric. 
    An incidence analysis is generated, with each value in the table reflecting the proportion of the population that 
    is above or below a threshold for a specified metric. There is an option to only provide a single `hrvar` in which a 
    bar plot is generated, or two `hrvar` values where an incidence table (heatmap) is generated.

    Parameters
    ----------
    data : pandas dataframe
        A Standard Person Query dataset in the form of a Pandas DataFrame.
    metric : str 
        Name of the metric, e.g. "Collaboration_hours".
    hrvar : str or list
         Name(s) of the HR Variable(s) by which to split metrics.
    mingroup : int
        Privacy threshold / minimum group size. Defaults to 5.
    threshold : float
        Threshold value to split the data based on the position argument. Defaults to None.
    position :  str 
        One of the below valid values:
        - "above": show incidence of those equal to or above the threshold
        - "below": show incidence of those equal to or below the threshold
    return_type : str
        What to return. This must be one of the following strings:
        - "plot"
        - "table"
                    
    Returns
    -------
    Output is returned depending on the value passed to the return_type argument:
    - "plot": Matplotlib or Seaborn plot object
    - "table": Pandas DataFrame
    
    Raises
    ------
    ValueError: If hrvar is not a string or list with at most length 2.

    Example
    -------
    """
    
    if not isinstance(hrvar, list):
        hrvar = [hrvar]
    if len(hrvar) > 2:
        raise ValueError("`hrvar` can only accept a list of length 2.")
    
    if len(hrvar) == 1:
        return create_inc_bar(data, metric, hrvar[0], mingroup, threshold, position, return_type)
    else:
        return create_inc_grid(data, metric, hrvar, mingroup, threshold, position, return_type)
    
def create_inc_bar(data: pd.DataFrame, metric: str, hrvar: str, mingroup: int = 5, threshold: float = None, position: str = None, return_type: str='plot'):
    """
    Name
    -----
    create_inc_bar

    Description
    -----------
    Run `create_inc` with only single `hrvar`. Returning a bar chart

    Parameters
    ----------
    data : pandas dataframe
        A Standard Person Query dataset in the form of a Pandas DataFrame.
    metric : str 
        Name of the metric, e.g. "Collaboration_hours".
    hrvar : str
        Name of the HR Variable by which to split metrics.
    mingroup : int
        Privacy threshold / minimum group size. Defaults to 5.
    threshold : float
        Threshold value to split the data based on the position argument. Defaults to None.
    position : str
        One of the below valid values:
        - "above": show incidence of those equal to or above the threshold
        - "below": show incidence of those equal to or below the threshold
    return_type : str 
        What to return. This must be one of the following strings:
        - "plot"
        - "table"
                    
    Returns
    -------
    Output is returned depending on the value passed to the return_type argument:
    - "plot": Matplotlib or Seaborn plot object
    - "table": Pandas DataFrame

    Raises
    ------
    ValueError: If hrvar is not a string.

    Example
    -------
    >>> create_inc_bar(data = pq_data, metric = "Collaboration_hours", hrvar = "LevelDesignation", theshold = 20, position = "below", return_type = "plot")
    """

    # Transform data so that metrics become proportions
    data_t = data.copy()
    if position == "above":
        data_t[metric] = data_t[metric] >= threshold
    elif position == "below":
        data_t[metric] = data_t[metric] <= threshold
    else:
        raise ValueError("Please enter a valid input for `position`.")
    
    
    title_text = f"Incidence of {metric} {position} {threshold}" # Set title text    
    subtitle_text = f"Percentage and number of employees by {hrvar}" # Set subtitle text
    
    return create_bar(
        data_t,
        metric,
        hrvar,
        mingroup,
        percent = True,
        plot_title = title_text,
        plot_subtitle = subtitle_text,
        return_type = return_type
        )

def create_inc_grid(data: pd.DataFrame, metric: str, hrvar: typing.List, mingroup: int=5, threshold: float=None, position: str=None, return_type: str='plot'):
    """
    Name
    -----
    create_inc_grid

    Description
    -----------
    Run `create_inc` with two `hrvar`.
    Returning a heatmap

    Parameters
    ----------
    data : pandas dataframe 
        A Standard Person Query dataset in the form of a Pandas DataFrame.
    metric : str
         Name of the metric, e.g. "Collaboration_hours".
    hrvar : list
         Names of the HR Variables by which to split metrics.
    mingroup : int
         Privacy threshold / minimum group size. Defaults to 5.
    threshold : float
         Threshold value to split the data based on the position argument. Defaults to None.
    position : str 
        One of the below valid values:
        - "above": show incidence of those equal to or above the threshold
        - "below": show incidence of those equal to or below the threshold
    return_type : str
        What to return. This must be one of the following strings:
        - "plot"
        - "table"
                    
    Returns
    -------
    Output is returned depending on the value passed to the return_type argument:
    - "plot": Matplotlib or Seaborn plot object
    - "table": Pandas DataFrame

    Raises
    ------
    ValueError: If hrvar is not a list of length 2.
    """

    if not isinstance(hrvar, list) or len(hrvar) != 2:
        raise ValueError("`hrvar` must be a list of length 2.")

    metric_to_pass = np.where(data[metric] >= threshold, 1, 0) \
        if position == "above" else np.where(data[metric] <= threshold, 1, 0) \
            if position == "below" else {}
    
    myTable: pd.DataFrame = (
        data
        .assign(metric_inc=metric_to_pass)
        .groupby(hrvar + ['PersonId'], as_index=False)
        .agg({'metric_inc': 'mean'})
        .groupby(hrvar, as_index=False)
        .agg({'metric_inc': 'mean', 'PersonId': 'nunique'})
        .rename(columns={'metric_inc': 'incidence', 'PersonId': 'count'})
        .query('count >= @mingroup')
        .sort_values('incidence', ascending=False)
        )

    if return_type == "table":
        
        return myTable

    elif return_type == "plot":

        # Set title text
        title_text = f"Incidence of {metric.replace('_', ' ').capitalize()} {position} {threshold}"

        # Set subtitle text
        subtitle_text = f"Percentage and number of employees by {hrvar[0]} and {hrvar[1]}"
        cap_str = extract_date_range(data, return_type = 'text')
        
        # Create the heatmap with the new annot DataFrame
        myTable['metric_text'] = myTable.apply(lambda row: f"{row['incidence']*100:.1f}% ({row['count']})", axis=1)
        
        # Order the columns and rows by the longest first to fit landscape plot
        if myTable[hrvar[0]].nunique() > myTable[hrvar[1]].nunique():
            hrvar = [hrvar[1], hrvar[0]]
            
        # Annotation to pass to heatmap    
        annot_df = myTable.pivot(index=hrvar[0], columns=hrvar[1], values='metric_text')
        
        # Setup plot size.
        fig, ax = plt.subplots(figsize=(7, 4))
        
        # Create grid 
        # Zorder tells it which layer to put it on. We are setting this to 1 and our data to 2 so the grid is behind the data.
        # ax.grid(which="major", axis='both', color='#758D99', alpha=0.6, zorder=1)        
        ax.grid(False)
        
        # Remove tick marks
        ax.tick_params(
            which='both',      # Both major and minor ticks are affected
            top=False,         # Remove ticks from the top
            bottom=False,      # Remove ticks from the bottom
            left=False,        # Remove ticks from the left
            right=False        # Remove ticks from the right
        )
        
        # plot heatmap
        sns.heatmap(
            myTable.pivot(index=hrvar[0], columns=hrvar[1], values='incidence'),
            annot = annot_df,
            fmt='',
            cmap=COLOR_PALLET_ALT_2,
            center=0.5,
            square=True,
            ax=ax
            )

        # Add in line and tag
        ax.plot(
            [0, .9], # Set width of line, previously [-0.08, .9]
            [0.9, 0.9], # Set height of line
            # [1.17, 1.17], # Set height of line
            transform = fig.transFigure, # Set location relative to plot
            clip_on = False,
            color = '#fe7f4f',
            linewidth = .6
        )

        ax.add_patch(
            plt.Rectangle(
                (0, 0.9), # Set location of rectangle by lower left corner, previously [-0.08, .9]
                0.05, # Width of rectangle
                -0.025, # Height of rectangle
                facecolor = '#fe7f4f',
                transform = fig.transFigure,
                clip_on = False,
                linewidth = 0
                )
        )        
        
        # Set title
        ax.text(
            x = 0, y = 1.00,
            s = title_text,
            transform = fig.transFigure,
            ha = 'left',
            fontsize = 13,
            weight = 'bold',
            alpha = .8
        )
        
        # Set subtitle
        ax.text(
            x = 0, y = 0.95,
            s = subtitle_text,
            transform = fig.transFigure,
            ha = 'left',
            fontsize = 11,        
            alpha = .8
        )
        
        # Set caption
        ax.text(x=0, y=0.02, s=cap_str, transform=fig.transFigure, ha='left', fontsize=9, alpha=.7)  
        
        # return the plot object
        return fig
        # plt.show()
           
        """ Legacy
        ax.set(title=title_text, xlabel=hrvar[1], ylabel=hrvar[0], aspect='equal')
        ax.text(1.1, 1.05, subtitle_text, transform=ax.transAxes, fontsize=14, va='center')        
        """
    else: 
        raise ValueError("Please enter a valid input for `return_type`: Either `table` or `plot`.")



import pandas as pd
import unittest
from vivainsights.create_boxplot import create_boxplot_viz
from vivainsights.create_boxplot import create_boxplot_calc
from vivainsights.create_boxplot import create_boxplot
from vivainsights.pq_data import load_pq_data
import matplotlib.pyplot as plt


class TestCreateBoxplotViz(unittest.TestCase):
    def test_create_boxplot_viz(self):
        # Load sample data
        pq_data = load_pq_data()
        
        # Create a box plot of the emails sent grouped by organization
        output = create_boxplot_viz(pq_data, 'Emails_sent', 'Organization', mingroup=1)
        
        # Check that the output is a matplotlib.figure.Figure object
        self.assertIsInstance(output, plt.Figure)



class TestCreateBoxplotCalc(unittest.TestCase):
    def test_create_boxplot_calc(self):
        # Load sample data
        pq_data = load_pq_data()
        
        # Calculate box plot data for the emails sent metric grouped by organization
        output = create_boxplot_calc(pq_data, 'Emails_sent', 'Organization', mingroup=1)
        
        # Check that the output is a pandas DataFrame
        self.assertIsInstance(output, pd.DataFrame)
        
        # Check that the DataFrame has the correct columns
        self.assertListEqual(output.columns.tolist(), ['PersonId', 'group', 'Emails_sent', 'Employee_Count'])
        


class TestCreateBoxplot(unittest.TestCase):
    def test_create_boxplot_data(self):
        # Load sample data
        pq_data = load_pq_data()
        
        # Create a box plot of the emails sent grouped by organization
        output = create_boxplot(pq_data, 'Emails_sent', 'Organization', mingroup=1, return_type='data')
        
        # Check that the output is a pandas DataFrame
        self.assertIsInstance(output, pd.DataFrame)
        
        # Check that the DataFrame has the correct columns
        self.assertListEqual(output.columns.tolist(), ['PersonId', 'group', 'Emails_sent', 'Employee_Count'])

    def test_create_boxplot_plot(self):
        # Load sample data
        pq_data = load_pq_data()
        
        # Create a box plot of the emails sent grouped by organization
        output = create_boxplot(pq_data, 'Emails_sent', 'Organization', mingroup=1, return_type='plot')
        
        # Check that the output is a matplotlib.figure.Figure object
        self.assertIsInstance(output, plt.Figure)

    def test_create_boxplot_table(self):
        #load sample data
        pq_data = load_pq_data()

        # Create a box plot of the emails sent grouped by organization
        output = create_boxplot(pq_data, 'Emails_sent', 'Organization', mingroup=1, return_type='table')

        # Check that the output is a pandas DataFrame
        self.assertIsInstance(output, pd.DataFrame)

        # Check that the DataFrame has the correct columns
        self.assertListEqual(output.columns.tolist(), ['index','group', 'mean', 'median', 'sd', 'min', 'max', 'n'])

        
if __name__ == '__main__':
    unittest.main()
        





import unittest
import pandas as pd
import matplotlib.pyplot as plt
import vivainsights as vi
import warnings

class TestIdentifyHolidayweeks(unittest.TestCase):
    def setUp(self):
        # set up pq_data
        self.pq_data = vi.load_pq_data()
        
    def test_return_type_text(self):
        with warnings.catch_warnings(record=True) as w:
            result = vi.identify_holidayweeks(data = self.pq_data, return_type="text")
            
            # Check if the result is a string
            self.assertIsInstance(result, str)        
        
            # Check if any warnings were generated
            self.assertLess(len(w), 1)
        
    def test_return_type_holidayweeks_data(self):
        result = vi.identify_holidayweeks(data=self.pq_data, return_type="holidayweeks_data")
        
        # Check if the result is a Pandas DataFrame
        self.assertIsInstance(result, pd.DataFrame)
        
    def test_return_type_data_cleaned(self):
        result = vi.identify_holidayweeks(data=self.pq_data, return_type="data_cleaned")
        
        # Check if the result is a Pandas DataFrame
        self.assertIsInstance(result, pd.DataFrame)
        
    def test_return_type_data_dirty(self):
        result = vi.identify_holidayweeks(data=self.pq_data, return_type="data_dirty")
        
        # Check if the result is a Pandas DataFrame
        self.assertIsInstance(result, pd.DataFrame)
        
    def test_return_type_plot(self):
        result = vi.identify_holidayweeks(data=self.pq_data, return_type="plot")
        
        # Check if the result is a Matplotlib plot
        self.assertIsInstance(result, plt.Figure)
            
            


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import pandas as pd
"""Extracts the date range from a dataframe."""

def extract_date_range(data: pd.DataFrame, return_type: str = "table"):
    """
    Name
    ----
    extract_date_range

    Description
    -----------
    The function `extract_date_range` extracts the date range from a dataframe and returns it either as
    a table or as a text string.

    Parameters
    ---------
    data : pandas dataframe
         The `data` parameter is a pandas DataFrame that contains the data from which you want to extract the date range. It should have at least one column that represents the date
    return_type : str
        The `return_type` parameter is a string that specifies the format in which the date range should be returned. It has two possible values:, defaults to table
    
    Returns 
    -------
    The function `extract_date_range` returns either a pandas DataFrame or a string, depending
    on the value of the `return_type` parameter.
    """
    date_var = None
    
    if "Date" in data.columns:
        date_var = pd.to_datetime(data["Date"], format="%m/%d/%Y")
    elif "MetricDate" in data.columns:
        date_var = pd.to_datetime(data["MetricDate"], format="%Y-%m-%d")
    elif all(x in data.columns for x in ["StartDate", "EndDate"]):
        date_var = pd.to_datetime(data[["StartDate", "EndDate"]].stack().reset_index(drop=True), format="%m/%d/%Y")

    if date_var is None:
        raise ValueError("Error: no date variable found.")
    """
    Data frame to output
    """
    outTable = pd.DataFrame({"Start": date_var.min(), "End": date_var.max()}, index=[0])

    if return_type == "table":
        return outTable
    elif return_type == "text":
        return f"Data from {outTable.iloc[0]['Start'].strftime('%Y-%m-%d')} to {outTable.iloc[0]['End'].strftime('%Y-%m-%d')}"


import unittest
import os
import vivainsights as vi
from vivainsights.pq_data import load_pq_data
from PIL import Image, ImageDraw
from vivainsights import export



class TestExport(unittest.TestCase):

    def setUp(self):
        # Create a temporary directory for test files
        self.test_dir = 'test_files'
        os.makedirs(self.test_dir, exist_ok=True)

    def tearDown(self):
        # Remove the temporary directory and its contents
        for file_name in os.listdir(self.test_dir):
            file_path = os.path.join(self.test_dir, file_name)
            os.remove(file_path)
        os.rmdir(self.test_dir)

    def test_generate_csv(self):
        pq_data=load_pq_data()
        file_path = os.path.join(self.test_dir)
        export(pq_data,
           file_format='csv',
           path=file_path,
           timestamp=False)
        # Check if the file was created
        self.assertTrue(os.path.exists(file_path))
        
        file_path=file_path+'.csv'
        # Check if the file is not empty
        with open(file_path, 'r') as file:
            content = file.read()
            self.assertNotEqual(content, '')\
                
                
    def test_generate_png(self):
        file_path = os.path.join(self.test_dir)
        pq_data=load_pq_data()
        trend=vi.create_trend(data = pq_data, metric = "Collaboration_hours", hrvar = "LevelDesignation")
        export(trend,
           file_format='png',
           path=file_path,
           timestamp=False)
        # Check if the file was created
        self.assertTrue(os.path.exists(file_path))
        
        
    def test_generate_svg(self):
        file_path = os.path.join(self.test_dir)
        pq_data=load_pq_data()
        trend=vi.create_trend(data = pq_data, metric = "Collaboration_hours", hrvar = "LevelDesignation")
        export(trend,
           file_format='svg',
           path=file_path,
           timestamp=False)
        # Check if the file was created
        self.assertTrue(os.path.exists(file_path))        

    def test_generate_jpeg(self):
        file_path = os.path.join(self.test_dir)
        pq_data=load_pq_data()
        trend=vi.create_trend(data = pq_data, metric = "Collaboration_hours", hrvar = "LevelDesignation")
        export(trend,
           file_format='jpeg',
           path=file_path,
           timestamp=False)
        # Check if the file was created
        self.assertTrue(os.path.exists(file_path))  


    def test_generate_pdf(self):
        file_path = os.path.join(self.test_dir)
        pq_data=load_pq_data()
        trend=vi.create_trend(data = pq_data, metric = "Collaboration_hours", hrvar = "LevelDesignation")
        export(trend,
           file_format='pdf',
           path=file_path,
           timestamp=False)
        # Check if the file was created
        self.assertTrue(os.path.exists(file_path))  

if __name__ == '__main__':
    unittest.main()


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
This module performs network analysis with a person-to-person query
"""
import vivainsights as vi
import pandas as pd
import igraph as ig
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.legend_handler import HandlerTuple
import matplotlib.lines as mlines
from matplotlib.backends.backend_pdf import PdfPages
import random
from sklearn.preprocessing import minmax_scale
import warnings
import time

def network_p2p(data, 
    hrvar = "Organization",
    return_type = "plot",
    centrality = None,
    community = None,
    weight = None,
    comm_args = None,
    layout = "mds",
    path = "",
    style = "igraph",
    bg_fill = "#FFFFFF",
    font_col = "grey20",
    legend_pos = "best",
    palette = "rainbow",
    node_alpha = 0.7,
    edge_alpha = 1,
    edge_col = "#777777",
    node_sizes = [1, 20],
    node_scale = 1,
    seed = 1,
    legend_ncols=0
):
    """
    Name
    ----
    network_p2p

    Description
    ------------
    This function returns a network plot given a data frame containing a person-to-person query.

    Parameters
    ----------
    data : dataframe 
        Data frame containing a person-to-person query.
    hrvar : str 
        String containing the label for the HR attribute.
    return_type : str 
        A different output is returned depending on the value passed to the `return_type` argument: 
        - `'plot'` (default)
        - `'plot-pdf'`
        - `'sankey'`
        - `'table'`
        - `'data'`
        - `'network'`

    centrality : str 
        string to determines which centrality measure is used to scale the size of the nodes. All centrality measures are automatically calculated when it is set to one of the below values, and reflected in the `'network'` and `'data'` outputs. 
        Measures include: 
        - `betweenness`
        - `closeness`
        - `degree`
        - `eigenvector`
        - `pagerank`
        When `centrality` is set to None, no centrality is calculated in the outputs and all the nodes would have the same size. 

    community : str 
        String determining which community detection algorithms to apply. Valid values include: 
        - `None` (default): compute analysis or visuals without computing communities.
        - `"multilevel"` (a version of louvain)
        - `"leiden"`
        - `"edge_betweenness"`
        - `"fastgreedy"`
        - `"infomap"`
        - `"label_propagation"`
        - `"leading_eigenvector"`
        - `"optimal_modularity"`
        - `"spinglass"`
        - `"walk_trap"`

    weight : str 
        String to specify which column to use as weights for the network. To create a graph without weights, supply `None` to this argument.
    comm_args : list
        list containing the arguments to be passed through to igraph's clustering algorithms. Arguments must be named. See examples section on how to supply arguments in a named list.
    layout : str
        String to specify the node placement algorithm to be used. Defaults to `"mds"` for the deterministic multi-dimensional scaling of nodes. 
        See <https://rdrr.io/cran/ggraph/man/layout_tbl_graph_igraph.html> for a full  list of options.
    path : str (file path)
        File path for saving the PDF output. Defaults to a timestamped path based on current parameters.
        
    bg_fill : str 
        String to specify background fill color.
    font_col : str
        String to specify font color.
    legend_pos : str
        String to specify position of legend. Valid values include:  
        String to specify position of legend. Valid values include: 
        - `"best"`
        - `"upper right"`
        - `"upper left"`
        - `"lower left"`
        - `"right"`
        -  `"center left"`
        - `"center right"`
        - `"lower center"`
        - `"upper center"`
        - `"center"`
    palette : str 
        String specifying the function to generate a color palette with a single argument `n`. Uses `"rainbow"` by default.
    node_alpha : int 
        A numeric value between 0 and 1 to specify the transparency of the nodes. Defaults to 0.7.
    :param edge_alpha : int
        A numeric value between 0 and 1 to specify the transparency of the edges (only for 'ggraph' mode). Defaults to 1.
    edge_col: String to specify edge link color.
    node_sizes: int
        Numeric vector of length two to specify the range of node sizes to rescale to, when `centrality` is set to a non-null value.
    node_scale: int
        A numeric value to multiply or divide the size of the nodes. 
        This is applied to the 'node_size' attribute in the graph to increase or decrease the size of the nodes.
    seed : int
        Seed for the random number generator passed to either `set.seed()` when the louvain or leiden community detection algorithm is used, to ensure consistency. Only applicable when `community` is set to one of the valid non-null values.
    legend_ncols : int
        Value is either 0 or 1, Parameter to change the orientation horizontal to vertical of legend in the plot.
    Returns
    -------
    A different output is returned depending on the value passed to the `return_type` argument:     
    - `'plot'`: return a network plot, interactively within R.
    - `'plot-pdf'`: save a network plot as PDF. This option is recommended when the graph is large, which make take a long time to run if `return_type = 'plot'` is selected. Use this together with `path` to control the save location.
    - `'sankey'`: return a sankey plot combining communities and HR attribute. This is only valid if a community detection method is selected at community`.
    - `'table'`: return a vertex summary table with counts in communities and HR attribute. When `centrality` is non-NULL, the average centrality values are calculated per group.
    - `'data'`: return a vertex data file that matches vertices with communities and HR attributes.
    - `'network'`: return 'igraph' object.

    Examples
    --------
    >>> vi.network_p2p(data = p2p_data, return_type = "plot")
    # Return a network visual
    
    >>> vi.network_p2p(data = p2p_data, community = "leiden", comm_args = {"resolution": 0.01}, return_type = "table")
    # Return the vertex table with counts in communities and HR attribute
    # Resolution is set to a low value to yield fewer communities
    
    >>> vi.network_p2p(data = p2p_data, centrality = "betweenness", return_type = "table")
    # Return the vertex table with centrality calculations
    
    >>> vi.network_p2p(
        data = p2p_data, # or whatever your query is stored
        node_scale = 50, # adjust this parameter to make nodes bigger/smaller
        return_type = "plot"
        )
    
    >>> vi.network_p2p(
        data=p2p_data, # or whatever your query is stored
        return_type = "sankey", # another return type for visualization 
        centrality = "betweenness", # centrality can be set as per requirement
        community = "leiden" # Adjust community 
        )
    # Return the sankey output based on centrality and community
    
    >>> vi.network_p2p(
        data=p2p_data, # or whatever your query is stored
        return_type = "plot",
        font_col = "grey20", # Color change option for fonts in chart
        legend_pos = "upper left", # Adjust the legend position using this parameter
        legend_ncols = 1 # Adjust this parameter to 0 or 1 to change legend orientation from vertical to horizontal
        )
    # Return the plot output based on different color scheme, legend orientation and position, font color change
    """
    path ="p2p" + ("" if community is None else '_' + community)
    
    # `style` is currently a placeholder as only igraph is supported
    # legacy argument from the R implementation
    style = "igraph"

    if len(node_sizes) != 2:
        raise ValueError("`node_sizes` must be of length 2")

    #Set data frame for edges
    if weight is None:
        edges = data.assign(NoWeight = 1).loc[:, ["PrimaryCollaborator_PersonId", "SecondaryCollaborator_PersonId", "NoWeight"]].rename(columns = {"NoWeight": "weight"})

    else:
        edges = data.loc[:, ["PrimaryCollaborator_PersonId", "SecondaryCollaborator_PersonId", weight]]

    pc_hrvar = "PrimaryCollaborator_" + hrvar
    sc_hrvar = "SecondaryCollaborator_" + hrvar

    # TieOrigin = PrimaryCollaborator
    tieOrigin = (
        edges[["PrimaryCollaborator_PersonId"]].drop_duplicates()
        .merge(data[["PrimaryCollaborator_PersonId", pc_hrvar]], on = "PrimaryCollaborator_PersonId", how = "left") #left join
        .rename(columns = {"PrimaryCollaborator_PersonId": "node"})
        .assign(**{hrvar: lambda row: row[pc_hrvar]}) #assign new column
        .drop(columns = [pc_hrvar]) 
    )

    # TieDest = SecondaryCollaborator
    tieDest = (
        edges[["SecondaryCollaborator_PersonId"]].drop_duplicates()
        .merge(data[["SecondaryCollaborator_PersonId", sc_hrvar]], on = "SecondaryCollaborator_PersonId", how = "left")
        .rename(columns = {"SecondaryCollaborator_PersonId": "node"})
        .assign(**{hrvar: lambda row: row[sc_hrvar]})
        .drop(columns = [sc_hrvar])
    )

    # Vertices data frame to provide meta-data
    vert_ft = pd.concat([tieOrigin, tieDest]).drop_duplicates()

    # Create igraph object
    g_raw = ig.Graph.TupleList(edges.itertuples(index=False), directed=True, weights=True)

    # Assign vertex attributes - HR attribute and node
    g_raw.vs[hrvar] = vert_ft[hrvar].tolist()
    g_raw.vs["node"] = vert_ft["node"].tolist()

    # Assign weights
    g_raw.es["weight"] = edges["weight"]

    # allowed community values
    valid_comm = ["leiden", "multilevel", "edge_betweenness", "fastgreedy", "infomap", "label_propagation", "leading_eigenvector", "optimal_modularity", "spinglass", "walk_trap"]

    # Finalise `g` object
    # If community detection is selected, this is where the communities are appended
    if community is None:
        
        # g = g_raw.simplify()
        g = g_raw # Note: NOT simplified as simplification may remove too many edges
        v_attr = hrvar 
        
    elif community in valid_comm:
        random.seed(seed)
        g_ud = g_raw.as_undirected() # Convert to undirected graph
        
        # combine arguments to clustering algorithms
        comm_func = getattr(ig.Graph, "community_" + community)
        if comm_args is None:
            comm_args = {}

        # call community detection function
        comm_out = comm_func(graph = g_ud, **comm_args)
        # g = g_ud.simplify()
        g = g_ud # Note: NOT simplified as simplification may remove too many edges
        g.vs["cluster"] = [str(member) for member in comm_out.membership]

        #Name of vertex attribute
        v_attr = "cluster"
    else: 
        raise ValueError("Please enter a valid input for `community`.")
        
    # centrality calculations ------------------------
    # valid values of `centrality`
    valid_cent = ['betweenness', 'closeness', 'degree', 'eigenvector', 'pagerank']  
    
    # attach centrality calculations if `centrality` is not None
    if centrality in valid_cent:
        g = vi.network_summary(g, return_type="network")
        node_sizes = (node_sizes[1] - node_sizes[0]) 
        node_sizes *= minmax_scale(g.vs[centrality]) + node_sizes #min and max values
        g.vs["node_size"] = node_sizes/100 #scale for plotting      
    elif centrality is None:
        # all nodes with the same size if centrality is not calculated
        # adjust for plotting formats
        if style == "igraph":
            g.vs["node_size"] = [0.08] * g.vcount()
        elif style == "ggraph":
            g.vs["node_size"] = [0.08] * g.vcount()
            node_sizes = [0.03,0.03] #fix node size
    else:
        raise ValueError("Please enter a valid input for `centrality`.")

    # Common area ------------------- ----------------
    # vertex table
    vert_ft = vert_ft.rename(columns = {"node": "name"})    
    
    if centrality is not None:
        if community is None:
            vert_tb = pd.DataFrame({
                "name": g.vs["name"],
                "betweenness": g.vs["betweenness"],
                "closeness": g.vs["closeness"],
                "degree": g.vs["degree"],
                "eigenvector": g.vs["eigenvector"],
                "pagerank": g.vs["pagerank"],
            })
        else :
            vert_tb = pd.DataFrame({
                "name": g.vs["name"],
                "cluster": g.vs[v_attr],
                "betweenness": g.vs["betweenness"],
                "closeness": g.vs["closeness"],
                "degree": g.vs["degree"],
                "eigenvector": g.vs["eigenvector"],
                "pagerank": g.vs["pagerank"],
            })
    else:
        if community is None:
            vert_tb = pd.DataFrame({
                "name": g.vs["name"],
            })
        else:
            vert_tb = pd.DataFrame({
                "name": g.vs["name"],
                "cluster": g.vs[v_attr]
            })

    vert_tb = vert_tb.merge(vert_ft, on = "name", how = "left").drop_duplicates() #merge hrvar to vertex table
    
    g_layout = g.layout(layout)

    out_path = path + '_' + time.strftime("%y%m%d_%H%M%S") + '.pdf'

    # Return outputs ---------------------------------------
    #use fast plotting method
    if return_type in ["plot", "plot-pdf"]:
        
        def rainbow(n):
            return [f"#{random.randint(0, 0xFFFFFF):06x}" for _ in range(n)]
            
        # Set colours
        vert_tb = vert_tb.drop_duplicates()
        
        colour_tb = (
            pd.DataFrame({v_attr: g.vs[v_attr]})
             .assign(colour = eval(f"{palette}(len(vert_tb))"))
        )

        # Colour vector
        colour_v = (
            pd.DataFrame({v_attr: g.vs[v_attr]})
            .merge(colour_tb, on = v_attr, how = "left")
            .loc[:, "colour"]
        )

        if style == "igraph":
            # Set graph plot colours
            # color_names = list(mcolors.CSS4_COLORS.keys())
            # g.vs["color"] = [mcolors.to_rgba(color_names[i % len(color_names)], alpha=node_alpha) for i in range(len(g.vs))]
    
            g.vs["frame_color"] = None
            g.es["width"] = 1

            #Internal basic plotting function used inside 'network_p2p()'
            def plot_basic_graph(lpos = legend_pos, pdf=False, node_scale=node_scale):
                
                fig, ax = plt.subplots(figsize=(10, 10))
                plt.rcParams["figure.facecolor"] = bg_fill
                layout_func = getattr(ig.Graph, f"layout_{layout}")
                
                # Get the unique values of the vertex attribute
                unique_values = list(set(g.vs[v_attr]))
                
                # Create a colormap with one color for each unique value
                cmap = mcolors.ListedColormap([plt.get_cmap('tab20')(i) for i in range(len(unique_values))])

                handles = []
                labels = []
                
                # Map each unique value to an index in the colormap
                value_to_index = {value: i for i, value in enumerate(unique_values)}
                
                # Legend
                for i, value in enumerate(unique_values):
                    marker = mlines.Line2D([0], [0], marker='o', color='w', label=value, markerfacecolor=cmap(i), markersize=5)
                    handles.append(marker)
                    labels.append(value)
                    
                # Set node colours
                for i, value in enumerate(g.vs[v_attr]):
                    index = value_to_index[g.vs[i][v_attr]]
                    color = cmap(index)
                    g.vs[i]["color"] = color

                g.vs["node_size"] = [x*node_scale for x in g.vs["node_size"]] # scale the size of the nodes
                
                ig.plot(
                    g,
                    layout = layout_func(g),
                    target=ax,
                    vertex_label = None,
                    vertex_size = g.vs["node_size"],
                    edge_arrow_mode = "0",
                    edge_arrow_size=0, 
                    edge_color = "#adadad",
                )              
                
                # Number of legend columns
                if legend_ncols==0:
                    if len(handles)<=10:
                        leg_cols=len(handles)
                    elif 10<len(handles)<=20:
                        leg_cols = (len(handles) // 2)
                    else:
                        leg_cols = (len(handles) // 4)
                        warnings.warn("There are over 20 unique node categories. Consider changing your grouping variable, merging existing groups, or tweaking algorithm parameters (if applicable).", UserWarning)
                else:
                    leg_cols=1
                    
                plt.legend(
                    handles = handles,
                    labels = labels,
                    handler_map={tuple: HandlerTuple(ndivide=20)},
                    loc = legend_pos,
                    edgecolor= edge_col,
                    frameon = True,
                    markerscale = 1,
                    fontsize= 5,
                    labelcolor = 'grey',
                    ncols = leg_cols
                )

                if pdf:
                    return fig
                
                return plt.show() #return 'ggplot' object

            # Default PDF output unless None supplied to path
            if return_type == "plot":
                
                plot_basic_graph(lpos = legend_pos)
                
            elif return_type == "plot-pdf":
                with PdfPages(out_path) as pdf:
                    pdf.savefig(plot_basic_graph(pdf=True))
                print(f"Saved to {out_path}.")

        else:
            raise ValueError("Invalid input for `style`.")
    
    elif return_type == "data":
        
        vert_tb = vert_tb.reset_index(drop = True) 
        
        return vert_tb
    
    elif return_type == "network":
        
        return g
    
    elif return_type == "sankey":
        if community is None:
            raise ValueError("Note: no sankey return option is available if `None` is selected at `community`. Please specify a valid community detection algorithm.")
        elif community in valid_comm:
            vi.create_sankey(data = vert_tb.groupby([hrvar, 'cluster']).size().reset_index(name='n'), var1=hrvar, var2='cluster')
    
    elif return_type == "table":
        if community is None:
            if centrality is None:
                vert_tb = vert_tb.groupby(hrvar).size().reset_index(name='n')
            else:
                vert_tb = vert_tb.groupby(hrvar).agg(
                    n=('betweenness', 'size'),
                    betweenness=('betweenness', 'mean'),
                    closeness=('closeness', 'mean'),
                    degree=('degree', 'mean'),
                    eigenvector=('eigenvector', 'mean'),
                    pagerank=('pagerank', 'mean')
                )
        elif community in valid_comm:
            if centrality is None:
                vert_tb = vert_tb.groupby([hrvar, 'cluster']).size().reset_index(name='n')
            else:
                vert_tb = vert_tb.groupby([hrvar, 'cluster']).agg(
                    n=('betweenness', 'size'),
                    betweenness=('betweenness', 'mean'),
                    closeness=('closeness', 'mean'),
                    degree=('degree', 'mean'),
                    eigenvector=('eigenvector', 'mean'),
                    pagerank=('pagerank', 'mean')
            )
                
        return vert_tb
            
    else:
        raise ValueError("invalid input for `return_type`.")


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
This module extracts HR attributes (organizational data) through a combination of detecting variable class, 
number of unique values, regular expressions. There is an option to return either just a list of the variable names 
or a DataFrame containing only the variables themselves.
"""
import pandas as pd
def extract_hr(
    data: pd.DataFrame,
    max_unique =50,
    exclude_constants =True,
    return_type: str = "names"):
    """    
    Name
    ----
    extract_hr

    Description
    -----------
    The function `extract_hr` extracts HR attributes (organizational data) through a combination of detecting variable class,

    Parameters
    ---------
    data : pandas dataframe 
        Contains the data to extract HR (highly-recurring) variables from
    max_unique: int 
        The maximum number of unique values a column can have to be included in the output, defaults to 50 (optional)
    exclude_constants: boolean
         A boolean value (True/False) indicating whether to exclude columns with constant values or not. If True, columns with constant values will be excluded. If False, all columns will be included regardless of whether they have constant values or not, defaults to True (optional)
    return_type : str
         The type of output to be returned, either "names" or "vars". If "names", the function will return the names of the columns that meet the specified criteria. If "vars", the function will return the actual columns that meet the specified criteria, defaults to names (optional)
    
    Returns 
    -------
    The function is not returning anything. It is printing the column names of the object columns in the filtered dataframe.
    """
    try:
        if((isinstance(max_unique, int)) and (isinstance(exclude_constants, bool))\
        and (return_type.lower() == "names") or (return_type.lower() == "vars")):
            unqdf = data.loc[:,data.nunique()<=max_unique]

        if exclude_constants == False:
            unqdf = unqdf.loc[:,unqdf.nunique()!=1]
  
        elif not isinstance(max_unique, int):
            error ="Error! var max_unique should be an integer value. Please try again."

        elif not isinstance(exclude_constants, bool):
            error ="Error! var exclude_constants should be an boolean(True/False) value. Please try again."

        elif (return_type.lower() != "names") or (return_type.lower() != "vars"):
            error ="Error! var return_type should be either(names/vars) value. Please try again."

        if return_type == "vars":
            return unqdf.select_dtypes(['object'])
        #return print(*unqdf.columns+',\n')
        return print(*unqdf.select_dtypes(['object']).columns+',\n')
    
    except:
        print(error)


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
""" 
This function scans a standard query output for weeks where collaboration
hours is far outside the mean. Returns a list of weeks that appear to be
holiday weeks and optionally an edited dataframe with outliers removed. By
default, missing values are excluded.
"""

import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedLocator


def identify_holidayweeks(data: pd.DataFrame, sd = 1, return_type = "text"):
    """"
    Name
    -----
    identify_holidayweeks

    Description
    -----------
    Identify Holiday Weeks based on outliers.
    This function scans a standard query output for weeks where collaboration
    hours is far outside the mean. Returns a list of weeks that appear to be
    holiday weeks and optionally an edited dataframe with outliers removed. By
    default, missing values are excluded.

    As best practice, run this function prior to any analysis to remove atypical
    collaboration weeks from your dataset.    

    Parameters
    ----------
    data : pandas dataframe
        A Standard Person Query dataset in the form of a data frame.
    sd : int 
        The standard deviation below the mean for collaboration hours that should 
        define an outlier week.  Enter a positive number. 
        Default is 1 standard deviation.
    return_type : str
        String specifying what to return. This must be one of the following strings:
        - "text" (default)
        - "labelled_data" or "dirty_data" or "data_dirty"
        - "cleaned_data" or "data_cleaned"
        - "holidayweeks_data"
        - "plot"

    Returns
    -------
    A different output is returned depending on the value passed to return_type:

    text : str
        A message is printed identifying holiday weeks.
    data_cleaned / cleaned_data : pandas dataframe
        A dataset with outlier weeks removed is returned.
    data_dirty / dirty_data / labelled_data : pandas dataframe
        A dataset with only outlier weeks is returned.
    holidayweeks_data : pandas dataframe
        A dataset with only outlier weeks is returned.
    plot : matplotlib plot 
        A line plot of Collaboration Hours with holiday weeks highlighted.

    Examples
    --------
    >>> identify_holidayweeks(pq_data, sd = .75, return_type = "text")
    "The weeks where collaboration was 0.75 standard deviations below the mean (18.7) are: `05/22/2022`"

    >>> identify_holidayweeks(pq_data, sd = .75, return_type = "plot")
    
    >>> identify_holidayweeks(pq_data, sd = .75, return_type = "cleaned_data")
    
    >>> identify_holidayweeks(pq_data, sd = .75, return_type = "holidayweeks_data")
    """

    try:
        # convert `MetricDate` to datetime
        data['MetricDate'] = pd.to_datetime(data.MetricDate)

        # Calculate the mean and z-score of collaboration hours by date
        Calc = data.dropna(subset=['MetricDate', 'Collaboration_hours']).groupby("MetricDate").agg(mean_collab = ("Collaboration_hours", "mean")).reset_index()
        Calc["z_score"] = (Calc["mean_collab"] - Calc["mean_collab"].mean())/ Calc["mean_collab"].std()


        # Find the outliers that are below the given standard deviation
        Outliers = (Calc["MetricDate"][Calc["z_score"] < -sd])

        Calc = Calc.assign(Outlier = Calc["MetricDate"].isin(Outliers))
        
        # Return the message or the plot depending on the argument
        if return_type== "text":
            # Calculate the total return_type and the message
            mean_collab_hrs = Calc["mean_collab"].mean()

            if len(Outliers) == 0:
                Message = 'There are no weeks where collaboration was ' + str(sd) + ' standard deviations below the mean (' + str(round(mean_collab_hrs, 1)) + ').'
            else:
                Message = 'The weeks where collaboration was ' + str(sd) + ' standard deviations below the mean (' + str(round(mean_collab_hrs, 1)) + ') are: '
                Message += ', '.join(Outliers.apply(lambda x: "`" + x.strftime("%m/%d/%Y") + "`"))
            
            return Message
        
        elif return_type in ["labelled_data", "dirty_data", "data_dirty"]:
            
            data_labelled = data.assign(holidayweek = data["MetricDate"].isin(Outliers))
            return data_labelled
        
        elif return_type == "cleaned_data" or return_type == "data_cleaned":
            # Calculate the three dataframe outputs
            data_cleaned = data[~data["MetricDate"].isin(Outliers)]
            return data_cleaned
        
        elif return_type == "holidayweeks_data":
            data_hw = data[data["MetricDate"].isin(Outliers)]
            return data_hw
        
        elif return_type == "plot":
            # Generate a line plot with matplotlib for the collaboration hours
            fig, ax = plt.subplots(figsize=(10, 6))

            # Plot the collaboration hours
            ax.plot(Calc["MetricDate"], Calc["mean_collab"].round(0), color="#1d627e", linewidth=3)

            # Add a marker to indicate the holiday weeks
            ax.scatter(Calc[Calc.Outlier==True]["MetricDate"], Calc[Calc.Outlier==True]["mean_collab"].round(0), color="#fe7f4f", marker="o", s=150, zorder=3)

            # Create the strings for the title, subtitle and caption
            subtitle_str = "Average collaboration hours where markers indicate holiday weeks"
            cap_str = "Data from week of {} to week of {}".format(Calc["MetricDate"].min().strftime("%b %d, '%y"), Calc["MetricDate"].max().strftime("%b %d, '%y"))

            # Set the title, subtitle, labels and limits of the plot
            ax.set_xlabel("Date", fontsize=12, fontweight="bold")
            ax.set_ylabel("Collaboration Hours", fontsize=12, fontweight="bold")
            ax.set_ylim(0, None)
            ax.text(x=ax.get_xlim()[0]-5, y=ax.get_ylim()[1]*1.10, s="Holiday Weeks", fontsize=16, fontweight="bold")
            ax.text(x=ax.get_xlim()[0]-5, y=ax.get_ylim()[1]*1.05, s=subtitle_str, fontsize=12)
            ax.text(x=ax.get_xlim()[0]-5,y=ax.get_ylim()[0]-5.5,s=cap_str, fontsize=12)
            
            ax.xaxis.set_major_locator(FixedLocator(range(len(Calc)))) # Set the tick positions
            ax.set_xticklabels(pd.to_datetime(Calc['MetricDate']).dt.strftime("%b %d, '%y"), rotation=45, ha="right")
            ax.grid(False)
            
            return fig
        else:
            raise ValueError("The `return_type` argument must be one of the following strings: 'text', 'labeled_data', 'cleaned_data', 'holidayweeks_data', or 'plot'.")
    except:

        # Check for the error in the input data
        required_cols = ['MetricDate', 'Collaboration_hours']
        for i in required_cols:
            if i not in data.columns:
                raise ValueError("The required variable {} is not present in the dataframe.".format(i))
                 
        if pd.api.types.is_datetime64_any_dtype(data['MetricDate'])==False:
            raise ValueError("`MetricDate` appears not to be properly formatted. It needs to be in the format YYYY-MM-DD. Also check for missing values or stray values with inconsistent formats.")
        



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
The function `totals_col` adds a new column with a specified total value to a given pandas DataFrame.
"""
import pandas as pd

def totals_col(data: pd.DataFrame, total_value: str ='Total'):
    '''    
    Name
    ----
    totals_col

    Description
    -----------
    The function `totals_col` adds a new column with a specified total value to a given pandas DataFrame.

    Parameters
    ----------
    data : pandas dataframe
        A pandas DataFrame that represents the data you want to add a totals column to.
    total_value : optional
        The `total_value` parameter is a string that represents the name of the new column that will be
    added to the DataFrame. By default, it is set to 'Total'.
    
    Returns
    -------
    The function `totals_col` returns the modified DataFrame `data` with a new column added.
    
    '''
    if total_value in data.columns:
        raise ValueError(f"Column '{total_value}' already exists. Please supply a different value to `total_value`")

    data[total_value] = total_value
    return data


import unittest
import matplotlib.pyplot as plt
from vivainsights.create_rank import create_rank_viz, create_rank_calc
from vivainsights.pq_data import load_pq_data
import tracemalloc
import gc
import pandas as pd

def returnfig():
    fig = plt.figure()
    return fig

class TestCreateRank(unittest.TestCase):
    def test_create_rank(self):
        tracemalloc.start()        
        pq_data = load_pq_data()        
        fig = create_rank_viz(data = pq_data,metric="Emails_sent", hrvar = ['Organization', 'FunctionType', 'LevelDesignation', 'SupervisorIndicator'],mingroup = 5)
        
        # garbage collection
        del pq_data
        gc.collect()
        
        tracemalloc.stop()
        self.assertIsInstance(fig, plt.Figure)
    
    def test_create_line_table(self):
        pq_data = load_pq_data()
        # Test if create_line returns a DataFrame for table output
        table = create_rank_calc(data=pq_data, metric="Emails_sent", hrvar=['Organization', 'FunctionType', 'LevelDesignation', 'SupervisorIndicator'], mingroup=5)
        self.assertIsInstance(table, pd.DataFrame)     

if __name__ == '__main__':
    unittest.main()